"""
The arcgis.mapping module provides components for visualizing GIS data and analysis.
This module includes WebMap and WebScene components that enable 2D and 3D
mapping and visualization in the GIS. This module also includes mapping layers like
MapImageLayer, SceneLayer and VectorTileLayer.
"""

from ._types import WebMap, WebScene, MapImageLayer, MapImageLayerManager, VectorTileLayer, OfflineMapAreaManager
from ._types import SceneLayer
from ._utils import export_map, get_layout_templates
from .symbol import create_symbol, display_colormaps, show_styles
from .renderer import generate_renderer
__all__ = ['WebMap', 'WebScene', 'MapImageLayer', 'MapImageLayerManager', 'VectorTileLayer',
           'export_map', 'get_layout_templates', 'OfflineMapAreaManager', 'SceneLayer']
