"""
All the spatial analysis tools from the analyze_patterns, enrich_data, find_locations, manage_data, summarize_data and
use_proximity submodules, in one place for convenience.
"""

from .analyze_patterns import *
from .use_proximity import *
from .enrich_data import *
from .find_locations import *
from .manage_data import *
from .summarize_data import *
